"""Public package for ih_muse."""

from ih_muse._utils.muse_version import get_muse_version as _get_muse_version
from ih_muse.config import ClientType, Config
from ih_muse.muse import Muse
from ih_muse.proto import (
    ElementKindRegistration,
    MetricDefinition,
    MetricPayload,
    MetricQuery,
    TimestampResolution,
)

__version__: str = _get_muse_version()
del _get_muse_version

__all__ = [
    "ClientType",
    "Config",
    "ElementKindRegistration",
    "MetricDefinition",
    "MetricPayload",
    "MetricQuery",
    "Muse",
    "TimestampResolution",
]



"""Module for configuring ih_muse."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from ih_muse.ih_muse import ClientType, PyConfig, TimestampResolution

if TYPE_CHECKING:
    from datetime import timedelta

    from ih_muse.proto import ElementKindRegistration, MetricDefinition


class Config:
    """Configuration class for the Muse client in Python.

    :param list[str] endpoints:
        List of endpoint URLs for the Muse client.
    :param ClientType client_type:
        The type of client to use (`ClientType.Poet` or `ClientType.Mock`).
    :param TimestampResolution default_resolution:
        Default timestamp resolution for metrics.
    :param list[ElementKindRegistration] element_kinds:
        List of element kinds to register.
    :param list[MetricDefinition] metric_definitions:
        List of metric definitions available for reporting.
    :param int max_reg_elem_retries:
        Maximum number of retries for element registration.
    :param bool recording_enabled:
        Enables event recording if set to `True`.
    :param Optional[str] recording_path:
        File path for recording events (required if `recording_enabled` is `True`).
    :param Optional[timedelta] recording_flush_interval:
        Interval to flush recordings (required if `recording_enabled` is `True`).
    :param Optional[timedelta] initialization_interval:
        Interval for the muse initialization task.
    :param Optional[timedelta] cluster_monitor_interval:
        Interval for the cluster monitoring task.

    ```python
    # Example usage:
    from ih_muse import Config, ClientType, TimestampResolution
    from ih_muse.proto import ElementKindRegistration, MetricDefinition

    config = Config(
        endpoints=["http://localhost:8080"],
        client_type=ClientType.Poet,
        default_resolution=TimestampResolution.Milliseconds,
        element_kinds=[ElementKindRegistration("kind_code", "description")],
        metric_definitions=[MetricDefinition("metric_code", "description")],
        max_reg_elem_retries=3,
        recording_enabled=False,
    )
    ```
    """

    _config: PyConfig

    def __init__(
        self,
        endpoints: list[str],
        client_type: ClientType,
        default_resolution: TimestampResolution,
        element_kinds: list[ElementKindRegistration],
        metric_definitions: list[MetricDefinition],
        max_reg_elem_retries: int,
        recording_enabled: bool,  # noqa: FBT001
        recording_path: Optional[str] = None,
        recording_flush_interval: Optional[timedelta] = None,
        initialization_interval: Optional[timedelta] = None,
        cluster_monitor_interval: Optional[timedelta] = None,
    ) -> None:
        """Initialize the Config instance."""
        py_element_kinds = [ekr._elem_kind_reg for ekr in element_kinds]
        py_metric_definitions = [md._metric_def for md in metric_definitions]

        self._config = PyConfig(
            endpoints,
            client_type,
            default_resolution,
            py_element_kinds,
            py_metric_definitions,
            max_reg_elem_retries,
            recording_enabled,
            recording_path,
            recording_flush_interval,
            initialization_interval,
            cluster_monitor_interval,
        )



"""Utilities for handling the Muse version."""

try:
    import ih_muse.ih_muse as muse

    _MUSE_VERSION = muse.__version__
except ImportError:
    # This is only useful for documentation
    import warnings

    warnings.warn("Polars binary is missing!", stacklevel=2)
    _MUSE_VERSION = ""


def get_muse_version() -> str:
    """Return the version of the Python IH-Muse package as a string.

    If the Muse binary is missing, returns an empty string.
    """
    return _MUSE_VERSION



"""Protocol definitions for ElementKind."""

from __future__ import annotations

from typing import Optional

from ih_muse.ih_muse import PyElementKindRegistration


class ElementKindRegistration:
    """Registration details for an ElementKind."""

    _elem_kind_reg: PyElementKindRegistration

    def __init__(
        self, code: str, name: str, description: str, parent_code: Optional[str] = None
    ) -> None:
        """Initialize ElementKindRegistration."""
        self._elem_kind_reg = PyElementKindRegistration(
            code, name, description, parent_code
        )



# ih_muse/exceptions.pyi

class MuseError(Exception): ...
class MuseInitializationTimeoutError(MuseError): ...
class ConfigurationError(MuseError): ...
class ClientError(MuseError): ...
class RecordingError(MuseError): ...
class ReplayingError(MuseError): ...
class InvalidFileExtensionError(MuseError): ...
class InvalidElementKindCodeError(MuseError): ...
class InvalidMetricCodeError(MuseError): ...
class DurationConversionError(MuseError): ...
class PanicException(MuseError): ...
class OtherPythonErrors(MuseError): ...



"""Exception definitions for ih_muse."""

try:
    from ih_muse.ih_muse import (
        ClientError,
        ConfigurationError,
        DurationConversionError,
        InvalidElementKindCodeError,
        InvalidFileExtensionError,
        InvalidMetricCodeError,
        MuseError,
        MuseInitializationTimeoutError,
        NotAvailableRemoteElementIdError,
        RecordingError,
        ReplayingError,
    )
except ImportError:
    # redefined for documentation purposes when there is no binary

    class MuseError(Exception):  # type: ignore[no-redef]
        """Base class for all IH-Muse errors."""

    class MuseInitializationTimeoutError(MuseError):  # type: ignore[no-redef, misc]
        """Exception raised when ...

        Examples
        --------
        >>> some code...
        ih_muse.exceptions.MuseInitializationTimeoutError: ...

        """

    class ConfigurationError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class ClientError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class RecordingError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class ReplayingError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class InvalidFileExtensionError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class InvalidElementKindCodeError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class NotAvailableRemoteElementIdError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class InvalidMetricCodeError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class DurationConversionError(MuseError):  # type: ignore[no-redef, misc]
        """TODO DOCS."""

    class PanicException(MuseError):  # type: ignore[no-redef, misc]
        """Exception raised on panics in the underlying Rust library."""


__all__ = [
    "ClientError",
    "ConfigurationError",
    "InvalidElementKindCodeError",
    "InvalidFileExtensionError",
    "InvalidMetricCodeError",
    "MuseError",
    "MuseInitializationTimeoutError",
    "NotAvailableRemoteElementIdError",
    "PanicException",
    "RecordingError",
    "ReplayingError",
]



# ih_muse/ih_muse.pyi

from datetime import timedelta
from typing import Any, Optional

__version__: str

class PyMuse:
    is_initialized: bool
    finest_resolution: TimestampResolution
    async def initialize(self, timeout: Optional[float] = None) -> None: ...
    def get_remote_element_id(self, local_elem_id: str) -> Optional[int]: ...
    async def register_element(
        self,
        kind_code: str,
        name: str,
        metadata: dict[str, str],
        parent_id: Optional[str] = None,
    ) -> str: ...
    async def send_metric(
        self,
        local_elem_id: str,
        metric_code: str,
        value: float,
    ) -> None: ...
    async def get_metrics(self, query: PyMetricQuery) -> list[PyMetricPayload]: ...
    async def replay(self, replay_path: str) -> None: ...
    def __init__(self, config: PyConfig) -> None: ...

class PyConfig:
    def __init__(
        self,
        endpoints: list[str],
        client_type: ClientType,
        default_resolution: TimestampResolution,
        element_kinds: list[Any],  # PyElementKindRegistration
        metric_definitions: list[Any],  # PyMetricDefinition
        max_reg_elem_retries: int,
        recording_enabled: bool,  # noqa: FBT001
        recording_path: Optional[str] = None,
        recording_flush_interval: Optional[timedelta] = None,
        initialization_interval: Optional[timedelta] = None,
        cluster_monitor_interval: Optional[timedelta] = None,
    ) -> None: ...

class ClientType:
    Poet: ClientType
    Mock: ClientType

class TimestampResolution:
    Years: TimestampResolution
    Months: TimestampResolution
    Weeks: TimestampResolution
    Days: TimestampResolution
    Hours: TimestampResolution
    Minutes: TimestampResolution
    Seconds: TimestampResolution
    Milliseconds: TimestampResolution
    Microseconds: TimestampResolution

    def to_timedelta(self) -> timedelta: ...

class PyMetricDefinition:
    def __init__(self, code: str, name: str, description: str) -> None: ...

class PyMetricPayload:
    time: int
    element_id: int
    metric_ids: list[int]
    values: list[Optional[float]]
    def __init__(
        self,
        time: int,
        element_id: int,
        metric_ids: list[int],
        values: list[Optional[float]],
    ) -> None: ...

class PyMetricQuery:
    def __init__(
        self,
        start_time: Optional[int],
        end_time: Optional[int],
        element_id: Optional[int],
        parent_id: Optional[int],
        metric_id: Optional[float],
    ) -> None: ...

class PyElementKindRegistration:
    def __init__(
        self, code: str, name: str, description: str, parent_code: Optional[str] = None
    ) -> None: ...



"""Public package for Protocol classes."""

from ih_muse.ih_muse import TimestampResolution

from .element_kind import ElementKindRegistration
from .metric import MetricDefinition, MetricPayload, MetricQuery

__all__ = [
    "ElementKindRegistration",
    "MetricDefinition",
    "MetricPayload",
    "MetricQuery",
    "TimestampResolution",
]



"""Protocol definitions for Metrics messages."""

from __future__ import annotations

from typing import Optional

from ih_muse.ih_muse import PyMetricDefinition, PyMetricPayload, PyMetricQuery


class MetricDefinition:
    """Definition of a metric."""

    _metric_def: PyMetricDefinition

    def __init__(self, code: str, name: str, description: str) -> None:
        """Initialize MetricDefinition."""
        self._metric_def = PyMetricDefinition(code, name, description)


class MetricPayload:
    """Definition of a metric payloadi."""

    _metric_payload: PyMetricPayload

    def __init__(
        self,
        time: int,
        element_id: int,
        metric_ids: list[int],
        values: list[Optional[float]],
    ) -> None:
        """Initialize MetricPayload."""
        self._metric_payload = PyMetricPayload(time, element_id, metric_ids, values)

    @classmethod
    def from_py_metric_payload(
        cls: type[MetricPayload], py_metric_payload: PyMetricPayload
    ) -> MetricPayload:
        """Create a MetricPayload from a PyMetricPayload.

        :param PyMetricPayload py_metric_payload:
            The PyMetricPayload to convert.

        :return:
            A corresponding MetricPayload instance.
        """
        return cls(
            time=py_metric_payload.time,
            element_id=py_metric_payload.element_id,
            metric_ids=py_metric_payload.metric_ids,
            values=py_metric_payload.values,
        )

    @property
    def time(self) -> int:
        """Retrieve the timestamp of the metric payload.

        :return:
            The time in milliseconds since epoch.
        """
        return self._metric_payload.time

    @property
    def element_id(self) -> int:
        """Retrieve the element ID associated with the metric payload.

        :return:
            The element ID as an integer.
        """
        return self._metric_payload.element_id

    @property
    def metric_ids(self) -> list[int]:
        """Retrieve the list of metric IDs.

        :return:
            A list of integers representing metric IDs.
        """
        return self._metric_payload.metric_ids

    @property
    def values(self) -> list[Optional[float]]:
        """Retrieve the list of metric values.

        :return:
            A list of floats or None for each metric value.
        """
        return self._metric_payload.values


class MetricQuery:
    """Definition of a metric."""

    _metric_query: PyMetricQuery

    def __init__(
        self,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        element_id: Optional[int] = None,
        parent_id: Optional[int] = None,
        metric_id: Optional[float] = None,
    ) -> None:
        """Initialize MetricQuery."""
        self._metric_query = PyMetricQuery(
            start_time, end_time, element_id, parent_id, metric_id
        )



"""Core functionality for Muse."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from ih_muse.ih_muse import PyMuse, TimestampResolution
from ih_muse.proto import MetricPayload

if TYPE_CHECKING:
    from ih_muse.config import Config
    from ih_muse.proto import MetricQuery


class Muse:
    """The main class for interacting with the Muse system in Python.

    :param Config config:
        The configuration object for initializing Muse.

    ```python
    # Example usage:
    from ih_muse import Muse, Config

    config = Config(...)
    muse = Muse(config)
    await muse.initialize(timeout=5.0)
    ```
    """

    _muse: PyMuse

    def __init__(self, config: Config) -> None:
        """Initialize the Muse instance."""
        self._muse = PyMuse(config._config)

    async def initialize(self, timeout: Optional[float] = None) -> None:
        """Initialize the Muse client and starts background tasks.

        :param Optional[float] timeout:
            Optional timeout in seconds for the initialization process.

        :raises MuseInitializationTimeoutError:
            If initialization times out.
        """
        await self._muse.initialize(timeout)

    @classmethod
    async def create(
        cls: type[Muse], config: Config, timeout: Optional[float] = None
    ) -> Muse:
        """Create and initialize a Muse instance.

        :param Config config:
            The configuration object.
        :param Optional[float] timeout:
            Optional timeout in seconds for initialization.

        :return:
            An initialized Muse instance.
        """
        instance = cls(config)
        await instance.initialize(timeout)
        return instance

    @property
    def is_initialized(self) -> bool:
        """Check whether the Muse client is initialized.

        :return:
            `True` if initialized, `False` otherwise.
        """
        return self._muse.is_initialized

    @property
    def finest_resolution(self) -> TimestampResolution:
        """Get the finest resolution stored by the Poet.

        :return:
            The current TimestampResolution stored in the poet.
        """
        return self._muse.finest_resolution

    def get_remote_element_id(self, local_elem_id: str) -> Optional[int]:
        """Retrieve the remote Element ID associated with a local Element ID.

        :param str local_elem_id:
            The local Element ID to query.

        :return:
            The corresponding remote Element ID, or `None` if not registered.
        """
        return self._muse.get_remote_element_id(local_elem_id)

    async def register_element(
        self,
        kind_code: str,
        name: str,
        metadata: dict[str, str],
        parent_id: Optional[str] = None,
    ) -> str:
        """Register a new element with the Muse system.

        :param str kind_code:
            The kind code of the element.
        :param str name:
            The name of the element.
        :param dict[str, str] metadata:
            Metadata associated with the element.
        :param Optional[int] parent_id:
            The parent element ID, if any.

        :return:
            The local element ID assigned to the registered element.

        :raises MuseError:
            If registration fails.
        """
        local_elem_id = await self._muse.register_element(
            kind_code,
            name,
            metadata,
            parent_id,
        )
        return local_elem_id

    async def send_metric(
        self,
        local_elem_id: str,
        metric_code: str,
        value: float,
    ) -> None:
        """Send a metric value associated with an element.

        :param int local_elem_id:
            The local ID of the element.
        :param str metric_code:
            The code identifying the metric.
        :param float value:
            The value of the metric to send.

        :raises MuseError:
            If sending the metric fails.
        """
        await self._muse.send_metric(
            local_elem_id,
            metric_code,
            value,
        )

    async def get_metrics(self, query: MetricQuery) -> list[MetricPayload]:
        """Retrieve metrics based on a query.

        :param MetricQuery query:
            The query specifying the criteria for retrieving metrics.

        :return:
            A list of MetricPayload matching the query.

        :raises MuseError:
            If retrieving the metrics fails.
        """
        raw_metrics = await self._muse.get_metrics(query._metric_query)
        return [MetricPayload.from_py_metric_payload(metric) for metric in raw_metrics]

    async def replay(self, replay_path: str) -> None:
        """Replays events from a recording file.

        :param str replay_path:
            The file path to the recording.

        :raises MuseError:
            If replaying the events fails.
        """
        await self._muse.replay(replay_path)
